#include "computation.h"
#include <iostream>
#include <vector>
#include <gmpxx.h>
using namespace std;

// XXX: this assumes int is 32 bits
// XXX: this assumes long long is 64 bits

#define COMPUTATION_TRUSTED 2 // input or already verified
#define COMPUTATION_ADD 3 // 2-input integer addition
#define COMPUTATION_SUB 4 // 2-input integer subtraction
#define COMPUTATION_MUL 5 // 2-input integer multiplication
#define COMPUTATION_NEGCMOV 20 // 3-input integer a+(a-b)c
#define COMPUTATION_NEGATE -1 // 1-input integer negation
#define COMPUTATION_MASK 11 // 1-input integer mask
#define COMPUTATION_LEFTSHIFT 26 // 1-input imm-distance integer left shift
#define COMPUTATION_RIGHTSHIFT 27 // 1-input imm-distance integer left shift
#define COMPUTATION_XOR 31 // 2-input integer xor
#define COMPUTATION_AND 32 // 2-input integer and
#define COMPUTATION_ANDFULL 33 // 2-input integer and, second input known to be in {-1,0}
#define COMPUTATION_INT32 -32 // reduce mod 2^32, signed
#define COMPUTATION_INT64 -64 // reduce mod 2^64, signed
#define COMPUTATION_UINT8 108 // reduce mod 2^8, unsigned
#define COMPUTATION_UINT32 132 // reduce mod 2^32 unsigned
#define COMPUTATION_UINT64 164 // reduce mod 2^64, unsigned

struct computation_node {
  mpz_class min;
  mpz_class max;

  long long op;
  long long source1;
  long long source2;
  long long source3;
  long long imm;

  int flagwanted; // temporary inside computation_assertequal
} ;

static vector<computation_node> computation = {
  {
    0, 0, COMPUTATION_TRUSTED, 0, 0, 0, 0, 0
  }
} ;

void computation_cut(long long id)
{
  computation[id].op = COMPUTATION_TRUSTED;
}

long long computation_const(int actual)
{
  if (actual == 0) return 0;

  computation_node n;
  long long result = computation.size();

  n.min = mpz_class(actual);
  n.max = n.min;
  n.op = COMPUTATION_TRUSTED;
  n.source1 = 0;
  n.source2 = 0;
  n.source3 = 0;
  n.imm = -1;

  cout << "# v" << result << " = ";
  cout << n.min;
  cout << " min " << n.min;
  cout << " max " << n.max;
  cout << "\n";

  computation.push_back(n);
  return result;
}

long long computation_inputint()
{
  computation_node n;
  long long result = computation.size();

  n.min = mpz_class("-2147483648");
  n.max = mpz_class("2147483647");
  n.op = COMPUTATION_TRUSTED;
  n.source1 = 0;
  n.source2 = 0;
  n.source3 = 0;
  n.imm = -1;

  computation.push_back(n);
  return result;
}

long long computation_inputunsignedchar()
{
  computation_node n;
  long long result = computation.size();

  n.min = 0;
  n.max = 255;
  n.op = COMPUTATION_TRUSTED;
  n.source1 = 0;
  n.source2 = 0;
  n.source3 = 0;
  n.imm = -1;

  computation.push_back(n);
  return result;
}

long long computation_add(long long source1,long long source2)
{
  computation_node n;
  long long result = computation.size();

  n.op = COMPUTATION_ADD;
  n.source1 = source1;
  n.source2 = source2;
  n.source3 = 0;
  n.imm = -1;

  n.min = computation[source1].min + computation[source2].min;
  n.max = computation[source1].max + computation[source2].max;

  cout << "# v" << result << " = ";
  cout << "v" << source1 << " + v" << source2;
  cout << " min " << n.min;
  cout << " max " << n.max;
  cout << "\n";

  computation.push_back(n);
  return result;
}

long long computation_negate(long long source1)
{
  computation_node n;
  long long result = computation.size();

  n.op = COMPUTATION_NEGATE;
  n.source1 = source1;
  n.source2 = 0;
  n.source3 = 0;
  n.imm = -1;

  n.min = -computation[source1].max;
  n.max = -computation[source1].min;

  cout << "# v" << result << " = ";
  cout << "-v" << source1;
  cout << " min " << n.min;
  cout << " max " << n.max;
  cout << "\n";

  computation.push_back(n);
  return result;
}

long long computation_xor(long long source1,long long source2)
{
  /* peephole: a^0 |-> a */
  if (computation[source2].min == 0)
    if (computation[source2].max == 0)
      return source1;

  /* peephole: 0^a |-> a */
  if (computation[source1].min == 0)
    if (computation[source1].max == 0)
      return source2;

  /* peephole: a^(a^b) |-> b */
  if (computation[source2].op == COMPUTATION_XOR)
    if (computation[source2].source1 == source1)
      return computation[source2].source2;
    
  computation_node n;
  long long result = computation.size();

  n.op = COMPUTATION_XOR;
  n.source1 = source1;
  n.source2 = source2;
  n.source3 = 0;
  n.imm = -1;

  if (computation[source1].min >= 0 && computation[source2].min >= 0) {
    n.min = 0;
    n.max = 0;
    for (;;) {
      if (computation[source1].max <= n.max)
        if (computation[source2].max <= n.max)
          break;
      n.max = n.max * 2 + 1;
    }
  } else {
    n.min = -1;
    n.max = 0;
    for (;;) {
      if (computation[source1].max <= n.max)
        if (computation[source2].max <= n.max)
          if (computation[source1].min >= n.min)
            if (computation[source2].max >= n.min)
              break;
      n.max = n.max * 2 + 1;
      n.min = n.min * 2;
    }
  }

  // a^((a^b)&c) is either a or b if c is known to be in {-1,0}
  // b^((a^b)&c) is either a or b if c is known to be in {-1,0}
  // a^(a&c) is either a or 0 if c is known to be in {-1,0}
  if (computation[source2].op == COMPUTATION_ANDFULL) {
    long long source22 = computation[source2].source2;
    long long source21 = computation[source2].source1;
    if (computation[source21].op == COMPUTATION_XOR) {
      if (computation[source21].source1 == source1) {
        long long source212 = computation[source21].source2;
        /* we have a^((a^b)&c) with c in {-1,0} with a=source1, b=source212 */
        /* i.e. a if c is 0, b if c is -1 */
        /* i.e. a+(a-b)c */
        /* i.e. negcmov(a,b,c) */
        mpz_class amin = computation[source1].min;
        mpz_class amax = computation[source1].max;
        mpz_class bmin = computation[source212].min;
        mpz_class bmax = computation[source212].max;
        n.min = amin;
        n.max = amax;
        if (bmin < n.min) n.min = bmin;
        if (bmax > n.max) n.max = bmax;
        n.op = COMPUTATION_NEGCMOV;
        n.source1 = source1;
        n.source2 = source212;
        n.source3 = source22;
      } else if (computation[source21].source2 == source1) {
        long long source211 = computation[source21].source1;
        /* we have b^((a^b)&c) with c in {-1,0} with a=source211, b=source1 */
        /* i.e. negcmov(b,a,c) */
        mpz_class amin = computation[source211].min;
        mpz_class amax = computation[source211].max;
        mpz_class bmin = computation[source1].min;
        mpz_class bmax = computation[source1].max;
        n.min = amin;
        n.max = amax;
        if (bmin < n.min) n.min = bmin;
        if (bmax > n.max) n.max = bmax;
        n.op = COMPUTATION_NEGCMOV;
        n.source1 = source1;
        n.source2 = source211;
        n.source3 = source22;
      }
    } else if (source21 == source1) {
      mpz_class amin = computation[source1].min;
      mpz_class amax = computation[source1].max;
      n.min = amin;
      n.max = amax;
      if (0 < n.min) n.min = 0;
      if (0 > n.max) n.max = 0;
      n.op = COMPUTATION_NEGCMOV;
      n.source1 = source1;
      n.source2 = 0;
      n.source3 = source22;
    }
  }

  cout << "# v" << result << " = ";
  cout << "v" << source1 << " ^ v" << source2;
  cout << " min " << n.min;
  cout << " max " << n.max;
  if (n.op == COMPUTATION_NEGCMOV) cout << " negcmov";
  cout << "\n";

  computation.push_back(n);
  return result;
}

long long computation_and(long long source1,long long source2)
{
  /* peephole: a & 0 |-> 0 */
  if (computation[source2].min == 0)
    if (computation[source2].max == 0)
      return 0;

  /* peephole: 0 & a |-> 0 */
  if (computation[source1].min == 0)
    if (computation[source1].max == 0)
      return 0;

  /* peephole: a & -1 |-> a */
  if (computation[source2].min == -1)
    if (computation[source2].max == -1)
      return source1;

  computation_node n;
  long long result = computation.size();

  n.op = COMPUTATION_AND;
  n.source1 = source1;
  n.source2 = source2;
  n.source3 = 0;
  n.imm = -1;

  // XXX: be more clever to narrow ranges
  if (computation[source1].min >= 0 && computation[source2].min >= 0) {
    n.min = 0;
    n.max = 0;
    for (;;) {
      if (computation[source1].max <= n.max)
        if (computation[source2].max <= n.max)
          break;
      n.max = n.max * 2 + 1;
    }
  } else {
    n.min = -1;
    n.max = 0;
    for (;;) {
      if (computation[source1].max <= n.max)
        if (computation[source2].max <= n.max)
          if (computation[source1].min >= n.min)
            if (computation[source2].max >= n.min)
              break;
      n.max = n.max * 2 + 1;
      n.min = n.min * 2;
    }
  }

  // a&c is either a or 0 if c is known to be in {-1,0}
  if (computation[source2].min >= -1)
    if (computation[source2].max <= 0) {
      n.op = COMPUTATION_ANDFULL;
      n.min = computation[source1].min;
      n.max = computation[source1].max;
      if (0 < n.min) n.min = 0;
      if (0 > n.max) n.max = 0;
    }

  cout << "# v" << result << " = ";
  cout << "v" << source1 << " & v" << source2;
  cout << " min " << n.min;
  cout << " max " << n.max;
  cout << "\n";

  computation.push_back(n);
  return result;
}

long long computation_mask(long long source1,long mask)
{
  computation_node n;
  long long result = computation.size();

  n.op = COMPUTATION_MASK;
  n.source1 = source1;
  n.source2 = 0;
  n.source3 = 0;
  n.imm = mask;

  // allow only power-of-2-minus-1 masks
  if (mask < 0) cout << "assert False\n";
  if (mask & (mask + 1)) cout << "assert False\n";

  n.min = 0;
  n.max = mask;

  cout << "# v" << result << " = ";
  cout << "v" << source1 << " & " << mask;
  cout << " min " << n.min;
  cout << " max " << n.max;
  cout << "\n";

  computation.push_back(n);
  return result;
}

long long computation_leftshift(long long source1,long long e)
{
  computation_node n;
  long long result = computation.size();

  n.op = COMPUTATION_LEFTSHIFT;
  n.source1 = source1;
  n.source2 = 0;
  n.source3 = 0;
  n.imm = e;

  n.min = computation[source1].min << e;
  n.max = computation[source1].max << e;

  cout << "# v" << result << " = ";
  cout << "v" << source1 << " << " << e;
  cout << " min " << n.min;
  cout << " max " << n.max;
  cout << "\n";

  computation.push_back(n);
  return result;
}

long long computation_rightshift(long long source1,long long e)
{
  computation_node n;
  long long result = computation.size();

  n.op = COMPUTATION_RIGHTSHIFT;
  n.source1 = source1;
  n.source2 = 0;
  n.source3 = 0;
  n.imm = e;

  n.min = computation[source1].min >> e;
  n.max = computation[source1].max >> e;

  cout << "# v" << result << " = ";
  cout << "v" << source1 << " >> " << e;
  cout << " min " << n.min;
  cout << " max " << n.max;
  cout << "\n";

  computation.push_back(n);
  return result;
}

static const mpz_class two8("256");
static const mpz_class two31("2147483648");
static const mpz_class two32("4294967296");
static const mpz_class two63("9223372036854775808");
static const mpz_class two64("18446744073709551616");

long long computation_uint8(long long source1)
{
  if (computation[source1].min >= 0)
    if (computation[source1].max <= two8 - 1)
      return source1;

  computation_node n;
  long long result = computation.size();

  n.op = COMPUTATION_UINT8;
  n.source1 = source1;
  n.source2 = 0;
  n.source3 = 0;
  n.imm = 0;

  mpz_class reducedmin = computation[source1].min % two8;
  if (reducedmin < 0) reducedmin += two8;
  mpz_class q = (computation[source1].min - reducedmin) / two8;
  // q is the multiple of 2^8 subtracted when reducing min

  if (computation[source1].max - q * two8 < two8) {
    n.min = reducedmin;
    n.max = computation[source1].max - q * two8;
  } else {
    n.min = 0;
    n.max = two8 - 1;
  }

  cout << "# v" << result << " = ";
  cout << " uint8 v" << source1;
  cout << " min " << n.min;
  cout << " max " << n.max;
  cout << "\n";

  computation.push_back(n);
  return result;
}

long long computation_uint32(long long source1)
{
  if (computation[source1].min >= 0)
    if (computation[source1].max <= two32 - 1)
      return source1;

  computation_node n;
  long long result = computation.size();

  n.op = COMPUTATION_UINT32;
  n.source1 = source1;
  n.source2 = 0;
  n.source3 = 0;
  n.imm = 0;

  mpz_class reducedmin = computation[source1].min % two32;
  if (reducedmin < 0) reducedmin += two32;
  mpz_class q = (computation[source1].min - reducedmin) / two32;
  // q is the multiple of 2^32 subtracted when reducing min

  if (computation[source1].max - q * two32 < two32) {
    n.min = reducedmin;
    n.max = computation[source1].max - q * two32;
  } else {
    n.min = 0;
    n.max = two32 - 1;
  }

  cout << "# v" << result << " = ";
  cout << " uint32 v" << source1;
  cout << " min " << n.min;
  cout << " max " << n.max;
  cout << "\n";

  computation.push_back(n);
  return result;
}

long long computation_uint64(long long source1)
{
  if (computation[source1].min >= 0)
    if (computation[source1].max <= two64 - 1)
      return source1;

  computation_node n;
  long long result = computation.size();

  n.op = COMPUTATION_UINT64;
  n.source1 = source1;
  n.source2 = 0;
  n.source3 = 0;
  n.imm = 0;

  mpz_class reducedmin = computation[source1].min % two64;
  if (reducedmin < 0) reducedmin += two64;
  mpz_class q = (computation[source1].min - reducedmin) / two64;
  // q is the multiple of 2^64 subtracted when reducing min

  if (computation[source1].max - q * two64 < two64) {
    n.min = reducedmin;
    n.max = computation[source1].max - q * two64;
  } else {
    n.min = 0;
    n.max = two64 - 1;
  }

  cout << "# v" << result << " = ";
  cout << " uint64 v" << source1;
  cout << " min " << n.min;
  cout << " max " << n.max;
  cout << "\n";

  computation.push_back(n);
  return result;
}

long long computation_int64(long long source1)
{
  if (computation[source1].min >= -two63)
    if (computation[source1].max <= two63 - 1)
      return source1;

  computation_node n;
  long long result = computation.size();

  n.op = COMPUTATION_INT64;
  n.source1 = source1;
  n.source2 = 0;
  n.source3 = 0;
  n.imm = 0;

  mpz_class reducedmin = computation[source1].min % two64;
  if (reducedmin < 0) reducedmin += two64;
  if (reducedmin >= two63) reducedmin -= two64;
  mpz_class q = (computation[source1].min - reducedmin) / two64;
  // q is the multiple of 2^64 subtracted when reducing min

  if (computation[source1].max - q * two64 < two63) {
    n.min = reducedmin;
    n.max = computation[source1].max - q * two64;
  } else {
    n.min = -two63;
    n.max = two63 - 1;
  }

  cout << "# v" << result << " = ";
  cout << " uint32 v" << source1;
  cout << " min " << n.min;
  cout << " max " << n.max;
  cout << "\n";

  computation.push_back(n);
  return result;
}

long long computation_int32(long long source1)
{
  if (computation[source1].min >= -two31)
    if (computation[source1].max <= two31 - 1)
      return source1;

  computation_node n;
  long long result = computation.size();

  n.op = COMPUTATION_INT32;
  n.source1 = source1;
  n.source2 = 0;
  n.source3 = 0;
  n.imm = 0;

  mpz_class reducedmin = computation[source1].min % two32;
  if (reducedmin < 0) reducedmin += two32;
  if (reducedmin >= two31) reducedmin -= two32;
  mpz_class q = (computation[source1].min - reducedmin) / two32;
  // q is the multiple of 2^32 subtracted when reducing min

  if (computation[source1].max - q * two32 < two31) {
    n.min = reducedmin;
    n.max = computation[source1].max - q * two32;
  } else {
    n.min = -two31;
    n.max = two31 - 1;
  }

  cout << "# v" << result << " = ";
  cout << " int32 v" << source1;
  cout << " min " << n.min;
  cout << " max " << n.max;
  cout << "\n";

  computation.push_back(n);
  return result;
}

long long computation_sub(long long source1,long long source2)
{
  computation_node n;
  long long result = computation.size();

  n.op = COMPUTATION_SUB;
  n.source1 = source1;
  n.source2 = source2;
  n.source3 = 0;
  n.imm = -1;

  n.min = computation[source1].min - computation[source2].max;
  n.max = computation[source1].max - computation[source2].min;

  // XXX: set up test cases for this
  if (computation[source2].op == COMPUTATION_LEFTSHIFT) {
    long long source21 = computation[source2].source1;
    if (computation[source21].op == COMPUTATION_RIGHTSHIFT) {
      if (computation[source2].imm == computation[source21].imm) {
        long long source211 = computation[source21].source1;
        if (source1 == source211) {
          // a - ((a >> e) << e)
          // XXX: don't narrow if not necessary
          n.min = 0;
          n.max = (mpz_class(1) << computation[source2].imm) - 1;
        } else if (computation[source211].op == COMPUTATION_ADD) {
          long long source2111 = computation[source211].source1;
          long long source2112 = computation[source211].source2;
          if (source1 == source2111) {
            if (computation[source2112].min == computation[source2112].max) {
              // a - (((a + c) >> e) << e)
              // XXX: don't narrow if not necessary
              n.min = 0;
              n.max = (mpz_class(1) << computation[source2].imm) - 1;
              n.min -= computation[source2112].max;
              n.max -= computation[source2112].min;
            }
          }
        }
      }
    }
  }

  cout << "# v" << result << " = ";
  cout << "v" << source1 << " - v" << source2;
  cout << " min " << n.min;
  cout << " max " << n.max;
  cout << "\n";

  computation.push_back(n);
  return result;
}

long long computation_mul(long long source1,long long source2)
{
  computation_node n;
  long long result = computation.size();

  n.op = COMPUTATION_MUL;
  n.source1 = source1;
  n.source2 = source2;
  n.source3 = 0;
  n.imm = -1;

  mpz_class possible;
  possible = computation[source1].min * computation[source2].min;
  n.min = possible;
  n.max = possible;
  possible = computation[source1].min * computation[source2].max;
  if (possible < n.min) n.min = possible;
  if (possible > n.max) n.max = possible;
  possible = computation[source1].max * computation[source2].min;
  if (possible < n.min) n.min = possible;
  if (possible > n.max) n.max = possible;
  possible = computation[source1].max * computation[source2].max;
  if (possible < n.min) n.min = possible;
  if (possible > n.max) n.max = possible;

  cout << "# v" << result << " = ";
  cout << "v" << source1 << " * v" << source2;
  cout << " min " << n.min;
  cout << " max " << n.max;
  cout << "\n";

  computation.push_back(n);
  return result;
}

int computation_knownpositive(long long source)
{
  return computation[source].min > 0;
}

int computation_knownnonpositive(long long source)
{
  return computation[source].max <= 0;
}

int computation_knownnegative(long long source)
{
  return computation[source].max < 0;
}

int computation_knownnonnegative(long long source)
{
  return computation[source].min >= 0;
}

void computation_assumemin(long long source,mpz_class min)
{
  if (computation[source].min < min) computation[source].min = min;
}

void computation_assumemax(long long source,mpz_class max)
{
  if (computation[source].max > max) computation[source].max = max;
}

void computation_assumerange(long long source,mpz_class min,mpz_class max)
{
  if (computation[source].min < min) computation[source].min = min;
  if (computation[source].max > max) computation[source].max = max;
}

void computation_assertrange(long long source,mpz_class min,mpz_class max)
{
  // XXX: think about how to report ranges to user
  // cout << "print " << computation[source].min << "," << computation[source].max << "\n";
  if (computation[source].min < min) cout << "assert False # " << computation[source].min << " < " << min << "\n";
  if (computation[source].max > max) cout << "assert False # " << computation[source].max << " > " << max << "\n";
}

void computation_assertconstant(long long source,mpz_class c)
{
  // XXX: think about how to report ranges to user
  if (computation[source].min != c) cout << "assert False\n";
  if (computation[source].max != c) cout << "assert False\n";
}

// assertion: x == a-(a-b)c; i.e. x is c ? b : a
void computation_assertcmov(long long x,long long a,long long b,long long c)
{
  if (computation[c].min < 0) cout << "assert False # control bit may be negative\n";
  if (computation[c].max > 1) cout << "assert False # control bit may be above 1\n";
  long long ab = computation_sub(a,b);
  long long abc = computation_mul(ab,c);
  long long y = computation_sub(a,abc);
  computation_assertequal(x,y);
}

// assertion: x,y are a,b swapped by condition bit c
void computation_assertcswap(long long x,long long y,long long a,long long b,long long c)
{
  computation_assertcmov(x,a,b,c);
  computation_assertcmov(y,b,a,c);
}

// assertion: c is a^b
void computation_assertxor(long long c,long long a,long long b)
{
  if (computation[c].op == COMPUTATION_XOR)
    if (computation[c].source1 == a)
      if (computation[c].source2 == b)
        return;

  if (computation[c].op == COMPUTATION_XOR)
    if (computation[c].source1 == b)
      if (computation[c].source2 == a)
        return;

  cout << "assert False # assertxor does not match\n";
}

// assertion: b is bit e of x
void computation_assertbit(long long b,long long x,long long e)
{
  // bit e of x1 + (x21 << c) is bit e-c of x21 if 0 <= x1 < 2^c and c<=e
  // bit e of x1 + (x21 << c) is bit e of x1 if e<c
  if (computation[x].op == COMPUTATION_ADD) {
    long long x1 = computation[x].source1;
    long long x2 = computation[x].source2;
    if (computation[x2].op == COMPUTATION_LEFTSHIFT) {
      long long c = computation[x2].imm;
      long long x21 = computation[x2].source1;
      if (c <= e)
        if (computation[x1].min >= 0)
          if (computation[x1].max < (mpz_class(1) << c)) {
            computation_assertbit(b,x21,e - c);
            return;
          }
      if (c > e) {
        computation_assertbit(b,x1,e);
        return;
      }
    }
  }

  // ok if b is (x >> e) & 1
  if (computation[b].op == COMPUTATION_MASK)
    if (computation[b].imm == 1) {
      long long b1 = computation[b].source1;
      if (computation[b1].op == COMPUTATION_RIGHTSHIFT)
        if (computation[b1].imm == e)
          if (computation[b1].source1 == x)
            return;
    }

  cout << "assert False # assertbit does not match\n";
}

void computation_assertequalmod(long long source1,long long source2,mpz_class mod)
{
  long long numvars = 0;

  for (long long i = 0;i < computation.size();++i) {
    computation[i].flagwanted = 0;
  }
  computation[source1].flagwanted = 1;
  computation[source2].flagwanted = 1;
  for (long long i = computation.size() - 1;i >= 0;--i) {
    if (computation[i].flagwanted && computation[i].op != COMPUTATION_TRUSTED) {
      computation[computation[i].source1].flagwanted = 1;
      computation[computation[i].source2].flagwanted = 1;
      computation[computation[i].source3].flagwanted = 1;
    }
  }

  cout << "R.<";
  for (long long i = 0;i < computation.size();++i) {
    if (!computation[i].flagwanted) continue;
    if (computation[i].op == COMPUTATION_TRUSTED)
      if (computation[i].min == computation[i].max) continue;
    switch(computation[i].op) {
      case COMPUTATION_ADD:
      case COMPUTATION_LEFTSHIFT:
      case COMPUTATION_NEGATE:
      case COMPUTATION_NEGCMOV:
      case COMPUTATION_ANDFULL:
      case COMPUTATION_SUB:
      case COMPUTATION_MUL:
        break;
      default:
        if (numvars) cout << ",";
        cout << "v" << i;
        ++numvars;
    }
  }
  if (numvars == 0) { // stupid Sage special cases
    cout << "vnothing";
    ++numvars;
  }
  cout << "> = PolynomialRing(ZZ," << numvars << ")\n";

  for (long long i = 0;i < computation.size();++i) {
    if (!computation[i].flagwanted) continue;
    if (computation[i].op == COMPUTATION_TRUSTED)
      if (computation[i].min == computation[i].max) {
        cout << "v" << i << " = " << computation[i].min << "\n";
        continue;
      }
    switch(computation[i].op) {
      case COMPUTATION_ADD:
        cout << "v" << i << " = v" << computation[i].source1 << " + v" << computation[i].source2 << "\n";
        break;
      case COMPUTATION_LEFTSHIFT:
        cout << "v" << i << " = v" << computation[i].source1 << " * 2^" << computation[i].imm << "\n";
        break;
      case COMPUTATION_NEGATE:
        cout << "v" << i << " = -v" << computation[i].source1 << "\n";
        break;
      case COMPUTATION_SUB:
        cout << "v" << i << " = v" << computation[i].source1 << " - v" << computation[i].source2 << "\n";
        break;
      case COMPUTATION_MUL:
        cout << "v" << i << " = v" << computation[i].source1 << " * v" << computation[i].source2 << "\n";
        break;
      case COMPUTATION_ANDFULL:
        cout << "v" << i << " = -(v" << computation[i].source1 << ") * v" << computation[i].source2 << "\n";
        break;
      case COMPUTATION_NEGCMOV:
        cout << "v" << i << " = v" << computation[i].source1 << " + (v" << computation[i].source1 << " - v" << computation[i].source2 << ") * v" << computation[i].source3 << "\n";
        break;
    }
  }

  cout << "I = (\n";
  cout << mod;
  cout << ") * R\n";

  cout << "assert v" << source1 << " - v" << source2 << " in I\n";
}

void computation_assertequal(long long source1,long long source2)
{
  computation_assertequalmod(source1, source2, 0);
  if (computation[source1].min < computation[source2].min) {
    computation[source1].min = computation[source2].min;
    cout << "# narrowing v" << source1 << " min " << computation[source1].min << "\n";
  }
  if (computation[source2].min < computation[source1].min) {
    computation[source2].min = computation[source1].min;
    cout << "# narrowing v" << source2 << " min " << computation[source2].min << "\n";
  }
  if (computation[source1].max > computation[source2].max) {
    computation[source1].max = computation[source2].max;
    cout << "# narrowing v" << source1 << " max " << computation[source1].max << "\n";
  }
  if (computation[source2].max > computation[source1].max) {
    computation[source2].max = computation[source1].max;
    cout << "# narrowing v" << source2 << " max " << computation[source2].max << "\n";
  }
}
